import asyncio
import websockets

from rocketchat_async.dispatcher import Dispatcher
from rocketchat_async.methods import Connect, Login, Resume, GetChannels, SendMessage,\
        SendReaction, SendTypingEvent, SubscribeToChannelMessages,\
        SubscribeToChannelChanges, Unsubscribe


class RocketChat:
    """Represents a connection to RocketChat, exposing the API."""

    class ConnectionClosed(Exception):
        pass

    class ConnectCallFailed(Exception):
        pass

    def __init__(self):
        self.user_id = None
        self.username = None
        self._dispatcher = Dispatcher(verbose=False)

    async def start(self, address, username, password):
        ws_connected = asyncio.get_event_loop().create_future()
        ws_connection = self._start(address, ws_connected)
        self._ws_connection_task = asyncio.create_task(ws_connection)
        try:
            await ws_connected
        except (OSError, websockets.InvalidMessage) as e:
            # Exceptions that can arise during temporary network glitches or
            # outages on the remote side.
            # See also https://github.com/aaugustin/websockets/issues/593
            raise self.ConnectCallFailed(e)

        # Connect and login.
        await self._connect()
        self.user_id = await self._login(username, password)
        self.username = username

    async def resume(self, address, username, token):
        ws_connected = asyncio.get_event_loop().create_future()
        ws_connection = self._start(address, ws_connected)
        self._ws_connection_task = asyncio.create_task(ws_connection)
        try:
            await ws_connected
        except (OSError, websockets.InvalidMessage) as e:
            # Exceptions that can arise during temporary network glitches or
            # outages on the remote side.
            # See also https://github.com/aaugustin/websockets/issues/593
            raise self.ConnectCallFailed(e)

        # Connect and login.
        await self._connect()
        self.user_id = await self._resume(token)
        self.username = username

    async def run_forever(self):
        try:
            await self.dispatch_task
        except websockets.ConnectionClosed as e:
            raise self.ConnectionClosed(e)

    async def _start(self, address, connected_fut):
        try:
            async with websockets.connect(address) as websocket:
                self.dispatch_task = self._dispatcher.run(websocket)
                # Notify the caller that login has succeeded.
                connected_fut.set_result(True)
                # Finally, create the ever-running dispatcher loop.
                await self.dispatch_task
        except Exception as e:
            if not connected_fut.done():
                connected_fut.set_exception(e)

    async def _connect(self):
        await Connect.call(self._dispatcher)

    async def _login(self, username, password):
        return await Login.call(self._dispatcher, username, password)

    async def _resume(self, token):
        return await Resume.call(self._dispatcher, token)

    # --> Public API methods start here. <--

    async def get_channels(self):
        """Get a list of channels user is currently member of."""
        return await GetChannels.call(self._dispatcher)

    async def send_message(self, text, channel_id, thread_id=None):
        """Send a text message to a channel."""
        await SendMessage.call(self._dispatcher, text, channel_id, thread_id)

    async def send_reaction(self, orig_msg_id, emoji):
        """Send a reaction to a specific message."""
        await SendReaction.call(self._dispatcher, orig_msg_id, emoji)

    async def send_typing_event(self, channel_id, thread_id=None):
        """Send the `typing` event to a channel."""
        await SendTypingEvent.call(self._dispatcher, channel_id, self.username, thread_id)

    async def subscribe_to_channel_messages(self, channel_id, callback):
        """
        Subscribe to all messages in the given channel.

        Returns the subscription ID.

        """
        sub_id = await SubscribeToChannelMessages.call(self._dispatcher,
                                                       channel_id, callback)
        return sub_id

    async def subscribe_to_channel_changes(self, callback):
        """
        Subscribe to all changes in channels.

        Returns the subscription ID.

        """
        sub_id = await SubscribeToChannelChanges.call(self._dispatcher,
                                                      self.user_id, callback)
        return sub_id

    async def unsubscribe(self, subscription_id):
        """Cancel a subscription."""
        await Unsubscribe.call(self._dispatcher, subscription_id)



import hashlib
import time


class RealtimeRequest:
    """Method call or subscription in the RocketChat realtime API."""
    _max_id = 0

    @staticmethod
    def _get_new_id():
        RealtimeRequest._max_id += 1
        return f'{RealtimeRequest._max_id}'


class Connect(RealtimeRequest):
    """Initialize the connection."""

    REQUEST_MSG = {
        'msg': 'connect',
        'version': '1',
        'support': ['1'],
    }

    @classmethod
    async def call(cls, dispatcher):
        await dispatcher.call_method(cls.REQUEST_MSG)


class Resume(RealtimeRequest):
    """Log in to the service with a token."""

    @staticmethod
    def _get_request_msg(msg_id, token):
        return {
            "msg": "method",
            "method": "login",
            "id": msg_id,
            "params": [
                {
                    "resume": token,
                }
            ]
        }

    @staticmethod
    def _parse(response):
        return response['result']['id']

    @classmethod
    async def call(cls, dispatcher, token):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, token)
        response = await dispatcher.call_method(msg, msg_id)
        return cls._parse(response)


class Login(RealtimeRequest):
    """Log in to the service."""

    @staticmethod
    def _get_request_msg(msg_id, username, password):
        pwd_digest = hashlib.sha256(password.encode()).hexdigest()
        return {
            "msg": "method",
            "method": "login",
            "id": msg_id,
            "params": [
                {
                    "user": {"username": username},
                    "password": {
                        "digest": pwd_digest,
                        "algorithm": "sha-256"
                    }
                }
            ]
        }

    @staticmethod
    def _parse(response):
        return response['result']['id']

    @classmethod
    async def call(cls, dispatcher, username, password):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, username, password)
        response = await dispatcher.call_method(msg, msg_id)
        return cls._parse(response)


class GetChannels(RealtimeRequest):
    """Get a list of channels user is currently member of."""

    @staticmethod
    def _get_request_msg(msg_id):
        return {
            'msg': 'method',
            'method': 'rooms/get',
            'id': msg_id,
            'params': [],
        }

    @staticmethod
    def _parse(response):
        # Return channel IDs and channel types.
        return [(r['_id'], r['t']) for r in response['result']]

    @classmethod
    async def call(cls, dispatcher):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id)
        response = await dispatcher.call_method(msg, msg_id)
        return cls._parse(response)


class SendMessage(RealtimeRequest):
    """Send a text message to a channel."""

    @staticmethod
    def _get_request_msg(msg_id, channel_id, msg_text, thread_id=None):
        id_seed = f'{msg_id}:{time.time()}'
        msg = {
            "msg": "method",
            "method": "sendMessage",
            "id": msg_id,
            "params": [
                {
                    "_id": hashlib.md5(id_seed.encode()).hexdigest()[:12],
                    "rid": channel_id,
                    "msg": msg_text
                }
            ]
        }
        if thread_id is not None:
            msg["params"][0]["tmid"] = thread_id
        return msg

    @classmethod
    async def call(cls, dispatcher, msg_text, channel_id, thread_id=None):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, channel_id, msg_text, thread_id)
        await dispatcher.call_method(msg, msg_id)


class SendReaction(RealtimeRequest):
    """Send a reaction to a specific message."""

    @staticmethod
    def _get_request_msg(msg_id, orig_msg_id, emoji):
        return {
            "msg": "method",
            "method": "setReaction",
            "id": msg_id,
            "params": [
                emoji,
                orig_msg_id,
            ]
        }

    @classmethod
    async def call(cls, dispatcher, orig_msg_id, emoji):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, orig_msg_id, emoji)
        await dispatcher.call_method(msg)


class SendTypingEvent(RealtimeRequest):
    """Send the `typing` event to a channel."""

    @staticmethod
    def _get_request_msg(msg_id, channel_id, username, thread_id=None):
        msg = {
            "msg": "method",
            "method": "stream-notify-room",
            "id": msg_id,
            "params": [
                f'{channel_id}/user-activity',
                username,
                ["user-typing"],
                {}
            ]
        }
        if(thread_id):
            msg["params"][-1]["tmid"] = thread_id
        return msg

    @classmethod
    async def call(cls, dispatcher, channel_id, username, thread_id=None):
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, channel_id, username, thread_id)
        await dispatcher.call_method(msg, msg_id)


class SubscribeToChannelMessages(RealtimeRequest):
    """Subscribe to all messages in the given channel."""

    @staticmethod
    def _get_request_msg(msg_id, channel_id):
        return {
            "msg": "sub",
            "id": msg_id,
            "name": "stream-room-messages",
            "params": [
                channel_id,
                {
                    "useCollection": False,
                    "args": []
                }
            ]
        }

    @staticmethod
    def _wrap(callback):
        def fn(msg):
            event = msg['fields']['args'][0]  # TODO: This looks suspicious.
            msg_id = event['_id']
            channel_id = event['rid']
            thread_id = event.get('tmid')
            sender_id = event['u']['_id']
            msg = event['msg']
            qualifier = event.get('t')
            unread = event.get('unread', False)
            repeated = bool(event.get('replies'))
            return callback(channel_id, sender_id, msg_id, thread_id, msg,
                            qualifier, unread, repeated)
        return fn

    @classmethod
    async def call(cls, dispatcher, channel_id, callback):
        # TODO: document the expected interface of the callback.
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, channel_id)
        await dispatcher.create_subscription(msg, msg_id, cls._wrap(callback))
        return msg_id  # Return the ID to allow for later unsubscription.


class SubscribeToChannelChanges(RealtimeRequest):
    """Subscribe to all changes in channels."""

    @staticmethod
    def _get_request_msg(msg_id, user_id):
        return {
            "msg": "sub",
            "id": msg_id,
            "name": "stream-notify-user",
            "params": [
                f'{user_id}/rooms-changed',
                False
            ]
        }

    @staticmethod
    def _wrap(callback):
        def fn(msg):
            payload = msg['fields']['args']
            if payload[0] == 'removed':
                return  # Nothing else to do - channel has just been deleted.
            channel_id = payload[1]['_id']
            channel_type = payload[1]['t']
            return callback(channel_id, channel_type)
        return fn

    @classmethod
    async def call(cls, dispatcher, user_id, callback):
        # TODO: document the expected interface of the callback.
        msg_id = cls._get_new_id()
        msg = cls._get_request_msg(msg_id, user_id)
        await dispatcher.create_subscription(msg, msg_id, cls._wrap(callback))
        return msg_id  # Return the ID to allow for later unsubscription.


class Unsubscribe(RealtimeRequest):
    """Cancel a subscription"""

    @staticmethod
    def _get_request_msg(subscription_id):
        return {
            "msg": "unsub",
            "id": subscription_id,
        }

    @classmethod
    async def call(cls, dispatcher, subscription_id):
        msg = cls._get_request_msg(subscription_id)
        await dispatcher.call_method(msg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rocketchat_async']

package_data = \
{'': ['*']}

install_requires = \
['websockets>=10.4,<11.0']

setup_kwargs = {
    'name': 'rocketchat-async',
    'version': '4.1.0',
    'description': 'asyncio-based Python wrapper for the Rocket.Chat Realtime API.',
    'long_description': '# rocketchat-async\n\nasyncio-based Python wrapper for the Rocket.Chat Realtime API.\n\nSupported Rocket.Chat versions: 6.x. (The library might also work or partially work with other versions.)\n\n## When should you use this library?\n\nUse this library if you:\n\n- want to integrate with Rocket.Chat from Python\n- are using [asyncio](https://docs.python.org/3/library/asyncio.html) to drive your code\n- want to use Rocket.Chat\'s efficient websockets-based Realtime API\n\n## Installation\n\n`pip install rocketchat-async`\n\n## Example usage\n\n```python\nimport asyncio\nimport random\nfrom rocketchat_async import RocketChat\n\n\ndef handle_message(channel_id, sender_id, msg_id, thread_id, msg, qualifier,\n                   unread, repeated):\n    """Simply print the message that arrived."""\n    print(msg)\n\n\nasync def main(address, username, password):\n    while True:\n        try:\n            rc = RocketChat()\n            await rc.start(address, username, password)\n            # Alternatively, use rc.resume for token-based authentication:\n            # await rc.resume(address, username, token)\n\n            # A possible workflow consists of two steps:\n            #\n            # 1. Set up the desired callbacks...\n            for channel_id, channel_type in await rc.get_channels():\n                await rc.subscribe_to_channel_messages(channel_id,\n                                                       handle_message)\n            # 2. ...and then simply wait for the registered events.\n            await rc.run_forever()\n        except (RocketChat.ConnectionClosed,\n                RocketChat.ConnectCallFailed) as e:\n            print(f\'Connection failed: {e}. Waiting a few seconds...\')\n            await asyncio.sleep(random.uniform(4, 8))\n            print(\'Reconnecting...\')\n\n\n# Side note: Don\'t forget to use the wss:// scheme when TLS is used.\nasyncio.run(main(\'ws://localhost:3000/websocket\', \'username\', \'password\'))\n```\n\n## API Overview\n\nBrief overview of the currently implemented methods.\n\nAs of now, Rocket.Chat\'s API is only covered partially (based on my original\nneeds). I am open to both feature requests as well as pull requests.\n\n### Methods\n\n#### `RocketChat.get_channels()`\n\nGet a list of channels the logged-in user is currently member of.\n\n#### `RocketChat.send_message(text, channel_id, thread_id=None)`\n\nSend a text message to a channel.\n\n#### `RocketChat.send_reaction(orig_msg_id, emoji)`\n\nSend a reaction to a specific message.\n\n#### `RocketChat.send_typing_event(channel_id, thread_id=None)`\n\nSend the "typing" event to a channel or to a specified thread within that channel.\n\n#### `RocketChat.subscribe_to_channel_messages(channel_id, callback)`\n\nSubscribe to all messages in the given channel. Returns the subscription ID.\n\nThe provided callback should accept eight arguments: `channel_id`,\n`sender_id`, `msg_id`, `thread_id`, `msg_text`, `msg_qualifier`\n and `repeated`. The qualifier can help to determine if e.g. the\nmessage is a system message about the user being removed from\nthe channel.  The `repeated` flag assists in distinguishing \nwhether the message has been received again as a result of \nthread replies, or if it is a new message post.\n\n#### `RocketChat.subscribe_to_channel_changes(callback)`\n\nSubscribe to all changes in channels. Returns the subscription ID.\n\nThe provided callback should accept two arguments: `channel_id`\nand `channel_qualifier`. The qualifier helps to determine e.g.\nif it\'s a direct message or a normal room.\n\n#### `RocketChat.unsubscribe(subscription_id)`\n\nCancel a subscription.\n',
    'author': 'Hynek Urban',
    'author_email': 'hynek.urban@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/hynek-urban/rocketchat-async',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import asyncio
import json
from typing import Any


class Dispatcher:
    """Match websockets calls with responses and manage callbacks."""

    def __init__(self, verbose=False):
        self._websocket: Any = None
        self._futures = {}  # ID -> asyncio Future (resolved with the response)
        self._verbose = verbose

        # ID -> registered callback (called when the msg arrives)
        self._callbacks = {}

    def run(self, websocket):
        self._websocket = websocket
        # Start listening to incoming messages, executing callbacks,
        # if appropriate.
        return asyncio.create_task(self._process_incoming())

    async def call_method(self, msg, msg_id=None):
        if self._verbose:
            print(f'Outgoing: {msg}')
        if (msg_id is None):
            await self._websocket.send(json.dumps(msg))
        else:
            fut = asyncio.get_event_loop().create_future()
            self._futures[msg_id] = fut
            await self._websocket.send(json.dumps(msg))
            return await fut

    async def create_subscription(self, msg, msg_id, callback):
        if self._verbose:
            print(f'Outgoing: {msg}')
        self._callbacks[msg['name']] = callback
        await self._websocket.send(json.dumps(msg))

    async def _process_incoming(self):
        try:
            while True:
                await self._process_incoming_event()
        except Exception as err:
            for fut in self._futures.values():
                # Propagate the exception to all awaiters to not get stuck.
                fut.set_exception(err)
            raise err

    async def _process_incoming_event(self):
        msg = await self._websocket.recv()
        if self._verbose:
            print(f'Incoming: {msg}')
        parsed = json.loads(msg)
        if parsed['msg'] == 'result':
            msg_id = parsed['id']
            if msg_id in self._futures:
                self._futures[msg_id].set_result(parsed)
                del self._futures[msg_id]
        elif parsed['msg'] == 'changed':  # Subscription update.
            stream_name = parsed['collection']
            if stream_name in self._callbacks:
                self._callbacks[stream_name](parsed)
        elif parsed['msg'] in ['ready', 'connected', 'added', 'updated',
                               'nosub']:
            return  # Nothing to do.
        elif parsed['msg'] == 'ping':
            asyncio.create_task(self.call_method({'msg': 'pong'}))
        elif parsed['msg'] == 'error':
            if self._verbose:
                print(f'Remote error: {msg}')
        else:
            if self._verbose:
                print(f'Unknown message: {msg}')



from enum import Enum


# Note: the mix-in classes should be eventually replaced by StrEnum.


class MessageQualifier(str, Enum):
    """Lists special message types."""
    REMOVE_USER = "ru"


class ChannelQualifier(str, Enum):
    """Lists special message types."""
    DIRECT_MESSAGE = "d"
    PRIVATE = "p"


class Emoji(str, Enum):
    """Selection of emojis that are subjectively "most fun"."""
    GRIN = ':grin:'
    SWEAT_SMILE = ':sweat_smile:'
    JOY = ':joy:'
    HEART_EYES = ':heart_eyes:'
    SMILING_FACE_WITH_3_HEARTS = ':smiling_face_with_3_hearts:'
    NERD = ':nerd:'
    SUNGLASSES = ':sunglasses:'
    PARTYING_FACE = ':partying_face:'
    SOB = ':sob:'
    EXPLODING_HEAD = ':exploding_head:'
    FEARFUL = ':fearful:'
    ROLLING_EYES = ':rolling_eyes:'
    THUMBSUP = ':thumbsup:'
    THUMBSDOWN = ':thumbsdown:'
    FINGERS_CROSSED = ':fingers_crossed:'
    METAL = ':metal:'
    V = ':v:'
    MAN_FACEPALMING = ':man_facepalming:'
    POINT_UP = ':point_up:'
    PENGUIN = ':penguin:'
